;;; -*- TI-Asm -*-
;;;
;;; Mimas - Assembly language IDE for the TI-83 Plus
;;; 
;;; Copyright (C) 2010-2012 Benjamin Moody
;;;
;;; This program is free software: you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;; ArchiveProgram:
;;
;; Save the current program to archive.  Throw an error if there is
;; insufficient archive space available.
;;
;; Destroys:
;; - AF, BC, DE, HL
;; - OP1, OP3

ArchiveProgram:
	;; if program is already archived, nothing to do
	ld a,(curProgStartPage)
	or a
	ret nz
	ld hl,curProgFileName
	call ArcUnarcVar
	;; fall through

;; ReloadProgram:
;;
;; Reload the current program, updating the current-program pointers
;; and curProgHeader.
;;
;; Destroys:
;; - AF, BC, DE, HL
;; - OP1
;; - ramCode

ReloadProgram:
	ld hl,curProgFileName
	rst rMOV9TOOP1
	;; fall through

;; LoadProgram:
;;
;; Load a new program.  Set up the current-program pointers, and copy
;; the program's header into RAM.
;;
;; Input:
;; - OP1 = program name
;;
;; Destroys:
;; - AF, BC, DE, HL
;; - OP1
;; - ramCode

LoadProgram:
	ld de,curProgFileName
	BCALL _MovFrOP1
	;; fall through

;; LoadTempProgram:
;;
;; Load a program without modifying curProgFileName.  The original
;; program can be reloaded by calling ReloadProgram.
;;
;; Input:
;; - OP1 = program name
;;
;; Destroys:
;; - AF, BC, DE, HL
;; - OP1
;; - ramCode

LoadTempProgram:
	call ChkFindSymVarDataStart
	jr c,LoadProgram_NotFound

	call Add_AHL_2		; skip program size
	ld (curProgStartPtr),hl
	ld (curProgStartPage),a

	;; Load program header
	ld de,curProgHeader
	ld bc,PROGHEADER_LENGTH
	push de
	 call FlashToRAM

	 ;; Calculate number of symbols
	 ld hl,(curProgHeader + PROGHEADER_SYMBOL_END)
	 ld de,(curProgHeader + PROGHEADER_SYMBOL_START)
	 or a
	 sbc hl,de
	 ld a,3
	 BCALL _DivHLByA
	 ld (curProgSymbolCount),hl

	 ;; Check if program signature is valid
	 pop de
	ld hl,sourceFileSignature
	ld b,5
	call CompStrsN
	ret z

  WARNING "FIXME: would be good to check that program header is sane"
;;; SYMBOL_START <= SYMBOL_STRING_START <= IMPORT_START <=
;;; SECTION_TABLE_START <= SECTION_TABLE_END <= size of program

LoadProgram_Invalid:
	ld hl,emsg_FileInvalid
	jr LoadProgram_Error

LoadProgram_NotFound:
	ld hl,emsg_FileNotFound
LoadProgram_Error:
	ld de,curProgFileName + 1
	call ThrowGeneralError
	;; UNREACHABLE

;; LoadFirstSection:
;;
;; Set up pointers and buffers for reading the first section of the
;; current input file.
;;
;; Output:
;; - Carry flag set if input file has no sections
;;
;; Destroys:
;; - AF, BC, DE, HL
;; - ramCode

LoadFirstSection:
	ld hl,(curProgHeader + PROGHEADER_SECTION_TABLE_START)
	jr LoadSectionAtOffset

;; LoadNextSection:
;;
;; Set up pointers and buffers for reading the next section of the
;; input file.
;;
;; Output:
;; - Carry flag set if no more sections to read
;;
;; Destroys:
;; - AF, BC, DE, HL
;; - ramCode

LoadNextSection:
	ld hl,(curSectionOffset)
	ld bc,SECTHEADER_LENGTH
	add hl,bc
LoadSectionAtOffset:
	ld de,(curProgHeader + PROGHEADER_SECTION_TABLE_END)
	or a
	sbc hl,de
	ccf
	ret c
	add hl,de
	ld (curSectionOffset),hl
	ld de,curSectionHeader
	ld bc,SECTHEADER_LENGTH
	;; fall through

;; CopyProgramData:
;;
;; Read data from the current input file.
;;
;; Input:
;; - HL = offset into program file
;; - DE = buffer to store data
;; - BC = number of bytes to copy
;;
;; Destroys:
;; - AF, BC, DE, HL
;; - ramCode

CopyProgramData:
	call ProgramOffsetToAddress
	jq_ FlashToRAM

;; GetProgramDataByte:
;;
;; Get a single byte from the current input file.
;;
;; Input:
;; - HL = program offset
;;
;; Output:
;; - A = byte
;;
;; Destroys:
;; - F
;; - ramCode

GetProgramDataByte:
	push hl
	 call ProgramOffsetToAddress
	 call Load_A_AHL
	 pop hl
	ret

;; GetProgramData:
;;
;; Get a chunk of data from the current input file.  If the
;; program is in Flash, the data will be copied to OP1.
;;
;; Input:
;; - HL = offset into program file
;; - BC = number of bytes to copy
;;
;; Output:
;; - HL points to data
;;
;; Destroys:
;; - AF, BC, DE, HL
;; - OP1-???
;; - ramCode

GetProgramData:
	call ProgramOffsetToAddress
	bit 7,h
	ret nz
	ld de,OP1
	push de
	 call FlashToRAM
	 pop hl
	ret

;; GetProgramInstr:
;;
;; Get a bytecode instruction from the current input file.  If the
;; program is in Flash, the string will be copied to instrBuf.
;;
;; Input:
;; - HL = program offset to instruction
;;
;; Output:
;; - HL points to instruction
;;
;; Destroys:
;; - AF, BC, DE, HL
;; - instrBuf
;; - ramCode

GetProgramInstr:
	call ProgramOffsetToAddress
	bit 7,h
	ret nz
	ld de,instrBuf
	push de
	 call InstrToRAM
	 pop hl
	ret

;; InstrToRAM:
;;
;; Copy a bytecode instruction to RAM.
;;
;; Input:
;; - AHL = paged address of instruction (4000h <= HL <= 7FFFh)
;; - DE = destination buffer
;;
;; Destroys:
;; - AF, BC, DE, HL
;; - ramCode

InstrToRAM:
  RAM_CODE_BEGIN InstrToRAM_end
	out (6),a
	ld c,(hl)
	res 7,c
	res 6,c
	inc c
	inc c
	ld b,0
InstrToRAM_Loop:
	ldi
	ret po
	bit 7,h
	jr z,InstrToRAM_Loop
	ld h,40h
	inc a
	out (6),a
	jr InstrToRAM_Loop
InstrToRAM_end:

;; GetProgramSymbolInfo:
;;
;; Get information about a symbol in the current program file.
;;
;; Input:
;; - HL = symbol number
;;
;; Output:
;; - Carry flag set if symbol does not exist
;; - HL = offset to the symbol's name from the start of the program
;;
;; Destroys:
;; - AF, DE

GetProgramSymbolInfo:
	call GetProgramSymbolOffset
	ret c
	call ProgramOffsetToAddress
	call Load_DE_AHL
	call Inc_AHL
	call Load_A_AHL
	cp 1
	ret c
;	ld a,d
;	res 6,d
	res 7,d
	ld hl,(curProgHeader + PROGHEADER_SYMBOL_STRING_START)
	add hl,de
	ret
