;;; -*- TI-Asm -*-
;;;
;;; Mimas plugin to display instruction info
;;;
;;; Copyright (C) 2012 Benjamin Moody
;;;
;;; This program is free software: you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

	processor z80
	nolist
	include <ti83plus.inc>
	list

MIMAS_PLUGIN equ 1

	nolist
	include <mimas.inc>
	include <defs.inc>
	include <insts.inc>
	include <plugapi.inc>
	list

;; PLUGIN_INFO:
;;
;; This plugin is called when the STAT key is pressed in the assembly
;; editor.  It parses and commits the current line, then (if the line
;; contains a valid instruction) displays a popup message with
;; information about that instruction.
;;
;; Input:
;; - Edit buffer contains the line being edited (might or might not
;;   have been modified)
;;
;; Destroys:
;; - AF, BC, DE, HL, IX
;; - instrBuf
;; - tempExprBuf
;; - OP1-OP6
;; - iMathPtr5
;; - saveSScreen

    BEGIN_PLUGIN PLUGIN_INFO

	call CommitAsmLine
	ret c
	call nz,ShowInstrInfo
	or a
	ret

;; ShowInstrInfo:
;;
;; Display a popup message with information about the
;; currently-selected instruction (size, speed, and affected CPU
;; flags.)  If the current line is not a Z80 instruction, return
;; without displaying anything.
;;
;; Input:
;; - (editLineOffset) = offset to selected instruction
;;
;; Destroys:
;; - AF, BC, DE, HL, IX
;; - instrBuf
;; - tempExprBuf
;; - OP1-OP6
;; - iMathPtr5
;; - saveSScreen

ShowInstrInfo:
	res editNewLine,(iy + asm_Flag3)

	;; set up environment so we can assemble a single instruction
	;; to see how large it is
	xor a
	ld (asmLoadPC), a	; don't care about MSB
	ld (flags + asm_Flag1),a
	inc a
	ld (asmPassNumber),a

	ld hl,(editLineOffset)
	call GetProgramInstr
	ld (iMathPtr5),hl

	;; check instruction type
	ld a,(hl)
	inc hl
	ld c,(hl)
	and 0C0h
	jr z,ShowInstrInfo_Normal
	cp T_SPECIAL
	ret nz
	;; special instruction: check if instr makes sense to analyze
	;; (most specials will just throw an error if we try to
	;; assemble them here; some will crash)
	ld a,c
	and 7Fh
	cp S_last + 1
	ret nc
	ld hl,instrInfoSpecialTable
	BCALL _IsInSet
	ret z
	ld c,0
ShowInstrInfo_Normal:
	push bc
	 push af
	  ld hl,(editLineOffset)
	  call AssembleProgramInstr

	  ;; Line 1: size of instruction
	  ld hl,instrInfoMsg
	  ld de,tempExprBuf
	  push de
	   pop ix
	  ld bc,instrInfoMsg_TStates - instrInfoMsg
	  ldir
	  ld a,(asmLoadPC)
	  call PutInstrInfoValue

	  pop af
	 pop bc
	jr nz,ShowInstrInfo_Done

	;; Line 2: number of T-states
	push bc
	 push hl
	  ld hl,normalInstructionTStateTable - pluginStart
	  call LookupInfoByte
	  pop hl

	 ld bc,instrInfoMsg_Flags - instrInfoMsg_TStates
	 ldir
	 ld (ix),3

	 cp 80h
	 jr c,ShowInstrInfo_IndirectHL
	 sub 0E0h
	 jr nc,ShowInstrInfo_TStatesNormal

	 ;; Conditional instruction
	 ;; encoded in T-state table as 1yy0xxxx
	 ;;   xxxx = base T-states minus 4
	 ;;     yy = extra T-states minus 5

	 push af
	  and 0Fh
	  add a,4
	  call PutInstrInfoValue
	  ld a,0BFh
	  ld (de),a
	  ld (ix+10),12
	  inc de
	  pop af
	 ld c,a
	 rlca
	 rlca
	 rlca
	 and 3
	 add a,c
	 and 1Fh
	 add a,8
ShowInstrInfo_TStatesDone:
	 call PutInstrInfoValue

	 ;; Line 3: list of affected flags
	 ld hl,instrInfoMsg_Flags
	 ld c,instrInfoMsg_End - instrInfoMsg_Flags
	 ldir
	 pop bc
	ld hl,normalInstructionFlagInfoTable - pluginStart
	call LookupInfoByte

	ld h,d
	ld l,e
	ld b,4
ShowInstrInfo_FlagsLoop:
	dec hl
	rrca
	jr c,ShowInstrInfo_FlagUsed
	ld (hl),0BDh
ShowInstrInfo_FlagUsed:
	djnz ShowInstrInfo_FlagsLoop

	;; Lines 4-5: details of flag usage
	and 7
	jr z,ShowInstrInfo_Done
	inc (ix)
	cp 2
	jr nc,ShowInstrInfo_OneInfoStr
	inc (ix)
ShowInstrInfo_OneInfoStr:
	ld hl,instrInfoFlagText
	ld c,b
ShowInstrInfo_FlagTextLoop:
	ld c,(hl)
	inc c
	add hl,bc
	dec a
	jr nz,ShowInstrInfo_FlagTextLoop
	ld c,23
	ldir
ShowInstrInfo_Done:
	ld hl,tempExprBuf
	ld de,instrInfoTitle
	jp ShowPopupMessage

ShowInstrInfo_IndirectHL:
	 ;; Instruction can take (HL) as an arg, which affects T-states
	 ;; - for all instrs except LD_r_n, the relevant arg is the last one
	 ;; - for LD_r_n, the relevant arg is the first one

	 ;; Encoded in T-state table as 0xxxyyyy
	 ;;    xxx = normal T-states minus 4
	 ;;   yyyy = indirect T-states minus 4

	 push de
	  ld d,a
	  ld hl,(iMathPtr5)
	  ld a,(hl)
	  and 3Fh
	  ld c,a
	  inc hl
	  ld a,(hl)
	  cp I_LD_r_n
	  jr nz,ShowInstrInfo_IndirectHLArg
	  ld c,1
ShowInstrInfo_IndirectHLArg:
	  add hl,bc
	  ld a,(hl)
	  cp X_REG_iHL
	  ld a,d
	  pop de
	 jr z,ShowInstrInfo_TStates4Bit
	 rrca
	 rrca
	 rrca
	 rrca
ShowInstrInfo_TStates4Bit:
	 and 0Fh
ShowInstrInfo_TStatesNormal:
	 add a,4
	 jr ShowInstrInfo_TStatesDone

LookupInfoByte:
	ld b,0
	add hl,bc
	push de
	 ex de,hl
	 ld hl,(myPluginStartPtr)
	 ld a,(myPluginStartPage)
	 call Add_AHL_DE
	 call Load_A_AHL
	 pop de
	ret

PutInstrInfoValue:
	push hl
	 ld l,a
	 ld h,0
	 BCALL _DivHLBy10
	 ld h,a
	 ld a,l
	 or a
	 call nz,PutInstrInfoValue1
	 ld a,0B4h
	 call z,PutInstrInfoChar
	 ld a,h
	 pop hl
PutInstrInfoValue1:
	xor 0CFh
PutInstrInfoChar:
	ld (de),a
	inc de
	ret

instrInfoSpecialTable:
	;;         ASCII   ASCIZ   HEX     INCBIN  BCALL   BJUMP   BREAK
	db         2 +     4 +     8 +     16 +    32 +    64 
	;; BREAK_c EQU     JQ_n    JQ_c_n  ALIGN   ASSERT  BLOCK   BYTE
	db                                                         128
	;; ELSE    ENDIF   IF      IFDEF   IFNDEF  ORG     RORG    WORD
	db                                                         128

instrInfoTitle:
	db 11, 0D9h,069h,0A8h,0FBh,0F2h,0A4h,076h,00Dh,096h,0F3h,070h ; Instruction Info

instrInfoMsg:
	db 1, 82
	db 7, 0ECh,04Eh,013h,0C5h,0B4h;,0B3h,0B3h ; Size:$##
instrInfoMsg_TStates:
	db 9, 0EBh,0ECh,0A1h,0A3h,09Ch,050h,0B4h;,0B3h,0B3h ; TStates: $##
instrInfoMsg_Flags:
instrInfoFlagText:
	db 10, 0D6h,051h,0F4h,09Ch,050h,0B4h,0ECh,0E5h,0EFh,0D3h ; Flags: $SZPC
instrInfoMsg_End:
	;; PARITY_C0 (2)
	db 8, 0B4h,0D3h,0D6h,00Fh,025h,031h,083h,020h ; $CF cleared
	;; PARITY (1)
	db 9, 0B4h,0EFh,0D6h,00Ch,020h,0F9h,018h,04Ah,0FFh ; $PF = parity
	;; OVERFLOW (1)
	db 11, 0B4h,0EFh,0D5h,00Ch,020h,07Fh,0C3h,08Fh,035h,07Fh,0D0h ; $PE = overflow
	;; INTSTATE (1)
	db 11, 0B4h,0EFh,0D5h,004h,0F3h,00Dh,09Dh,06Dh,060h,093h,0A0h ; $PE if IFF set
	;; BLOCK (1)
	db 10, 0B4h,0EFh,0DFh,004h,0F3h,00Dh,02Dh,03Ch,02Ch,0F0h ; $PO if BC=0
	;; BLOCKIO (2)
	db 8, 0B4h,0E5h,004h,0F3h,00Dh,02Ch,02Ch,0F0h ; $Z if B=0

    END_PLUGIN

;;; Info tables (note that these are not included in the plugin proper
;;; due to their size)

IC_NONE         equ 00h
IC_PARITY_C0    equ 10h
IC_PARITY       equ 20h
IC_OVERFLOW     equ 30h
IC_INTSTATE     equ 40h
IC_BLOCK        equ 50h
IC_BLOCKIO      equ 60h

	include "iinfo.asm"
